# -*- coding: utf-8 -*-

import pandas

from base_parser_class import InteropBinParser

from utils import dmesg

class InteropTileMetrics(InteropBinParser):
    "ILMN Tile Metrics parser (child class of InteropBinParser)."

    __version = 0.4                 # version of this parser class.
    supported_versions = [2]        # version(s) of binary file that this parser handles

    # given by __init__ (from InteropBinParser):  read_config {}, flowcell_layout {}

    def _init_variables(self):
        # Map of binary's "code" to what it means. Filled during parsing.
        self.codemap = { }
    
        #filled during parsing. 'code' refers to the binary's arbitrary outcome codes for each record.

        self.data = {'lane' : [], 
                     'tile': [], 
                     'code': [], 
                     'value': [] }

        # per-read calculations of average phasing and prephasing across all tiles.
        # index reads (usually Read 2) almost always report 0.0 phasing and prephasing.
        
        self.mean_phasing = []
        self.mean_prephasing = []
        
    def _make_codemap(self):
        self.codemap = { 100: "cluster density (k/mm2)",
                101: "cluster density passing filters (k/mm2)",
                102: "number of clusters",
                103: "number of clusters passing filters",
                400: "control lane" }
        
        for read in self.read_config:
            self.codemap[200 + (read['read_num']-1) * 2] = "phasing for read %i" % read['read_num']
            self.codemap[201 + (read['read_num']-1) * 2] = "prephasing for read %i" % read['read_num']
            self.codemap[300 + read['read_num']-1] = "percent aligned for read %i" % read['read_num']

    def _get_mean_of_last_cycle(self, df):
        if df.empty:
            return 0
        else:
            return df[len(df)-self.num_tiles:].mean()['value']

    def parse_binary(self):
        "parses contents of TileMetrics.bin / TileMetricsOut.bin, file version 2."

        # Contains aggregate or read metrics by tile
        # Format:
        #   byte 0: file version number (2)
        #   byte 1: length of each record
        #   bytes (N * 10 + 2) - (N *10 + 11): record:      
        #       2 bytes: lane number (uint16)
        #       2 bytes: tile number (uint16)
        #       2 bytes: metric code (uint16)
        #       4 bytes: metric value (float)
        # Where N is the record index and possible metric codes are:
        #   code 100: cluster density (k/mm2)
        #   code 101: cluster density passing filters (k/mm2)
        #   code 102: number of clusters
        #   code 103: number of clusters passing filters
        #   code (200 + (N – 1) * 2): phasing for read N
        #   code (201 + (N – 1) * 2): prephasing for read N
        #   code (300 + N – 1): percent aligned for read N
        #   code 400: control lane

        bs = self.bs    #convenience assignment
    
        self.apparent_file_version = bs.read('uintle:8')
    
        self.check_version(self.apparent_file_version)

        recordlen = bs.read('uintle:8')  # length of each record == 10
        bs.pos = 16  # skip the following 16 bytes which are invariant here.

        #read records bytewise per specs in technote_rta_theory_operations.pdf from ILMN
        for i in range(0,((bs.len - 16) / (recordlen * 8))):  # 80 == record length in bits
            self.data['lane'].append(bs.read('uintle:16'))          #lane number
            self.data['tile'].append(bs.read('uintle:16'))          #tile number
            self.data['code'].append(bs.read('uintle:16'))          #metric code
            self.data['value'].append(bs.read('floatle:32'))        #metric value

        #make it fuzzy and mean.
        self.df = pandas.DataFrame(self.data)

        # INTERPRETATION: MOVE TO SEPARATE FUNCTION(S)
            
        pivot_sum = self.df.pivot_table('value', rows='code', aggfunc='sum')
        pivot_mean = self.df.pivot_table('value', rows='code', aggfunc='mean')
        
        try:
            self.total_cluster_density = pivot_sum[100]
        except:
            self.total_cluster_density = 0    
    
        try:
            self.total_cluster_density_pf = pivot_sum[101]
        except:
            self.total_cluster_density_pf = 0    

        # SAV: "Total Reads"  
        # ResequencingRunStatistics.xml: NumberOfClustersRaw
        try:
            self.num_clusters = pivot_sum[102]      
        except:
            self.num_clusters = 0

        # SAV: "PF Reads"  
        try:                                        
            self.num_clusters_pf = pivot_sum[103]
        except:
            self.num_clusters_pf = 0             
                                              
        # Illumina SAV displays metrics only based on the latest-created cluster density (100)
        # and cluster density passing filter (101) metrics output per tile. (The number of collections
        # of tile metrics per sequencing run seems to be variable.)  So we select out the highest-index
        # metrics encompassing all num_tiles tiles, and calculate our means based on that.
    
        self.mean_cluster_density = self._get_mean_of_last_cycle(self.df[self.df['code']==100])        
        self.mean_cluster_density_pf = self._get_mean_of_last_cycle(self.df[self.df['code']==101])

        if self.num_clusters and self.num_clusters_pf:
            self.percent_pf_clusters = 100 * float(self.num_clusters_pf / self.num_clusters)
        else:
            self.percent_pf_clusters = 0
    
        # Phasing and Prephasing averages per read
        for read in self.read_config:
            # There are only ever (lanes * tiles) entries per phasing and pre-phasing code, so 
            # we don't need to do the "last cycle" trick as above.
            try:    
                self.mean_phasing.append(pivot_mean[200 + (read['read_num']-1) * 2])
                self.mean_prephasing.append(pivot_mean[201 + (read['read_num']-1) * 2])
            except:
                self.mean_phasing.append(0)
                self.mean_prephasing.append(0)

    def __str__(self):
        out = '  Mean Cluster Density: %i' % self.mean_cluster_density
        out += '\n  Mean PF Clusters Density: %i' % self.mean_cluster_density_pf
        out += '\n  Total Clusters: %i' % self.num_clusters
        out += '\n  Total PF Clusters: %i' % self.num_clusters_pf 
        out += '\n  Percentage of Clusters PF: %f' % self.percent_pf_clusters
        out += '\n  Read - PHASING / PRE-PHASING:'
        for read_num in range(self.num_reads):
            out += '\n    %i - %f / %f' % (read_num+1, self.mean_phasing[read_num], self.mean_prephasing[read_num])
        out += '\n'
        return out

    def to_dict(self):
        return { 'cluster_density': self.mean_cluster_density, 
                 'cluster_density_pf': self.mean_cluster_density_pf,
                 'num_clusters_pf': self.num_clusters_pf,
                 'num_clusters': self.num_clusters,
                 'mean_phasing': self.mean_phasing,
                 'mean_prephasing': self.mean_prephasing }


if __name__=='__main__':
    
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print 'supply path to TileMetrics.bin (or TileMetricsOut.bin)'
        sys.exit()
    
    TM = InteropTileMetrics(filename)
    
    print 'Length of data: %i' % len(TM.data['code'])
    #print TM.df.head()
    
    #print TM.to_dict()
    
    print TM
    



# -*- coding: utf-8 -*-

import pandas

from base_parser_class import InteropBinParser
from utils import dmesg, set_column_sequence

class InteropQualityMetrics(InteropBinParser):
    "ILMN Quality metrics parser (child class of InteropBinParser)."

    __version = 0.2     # version of this parser class.
    supported_versions = [4]        # version(s) of file that this parser supports

    # a scalar representing the number of quality scores per record in QualityMetrics*.bin
    num_quality_scores = 50

    def _init_variables(self):
        dmesg("QM Read Config:", 2)
        dmesg(self.read_config, 2)
        self._setup_read_tiers()
        
        # create .qcol_sequence to retain correct order of qx columns
        self._set_qcol_sequence()
        
        # create data dict for parser, making use of .qcol_sequence
        self._setup_data()

        # Q30 / Q20 scores per read, calculated at end of parsing.
        self.read_qscore_results = {'readnum': [], 'q30': [], 'q20': [] }

    def _set_qcol_sequence(self):
        "sets .qcol_sequence array of q score labels in correct order."
        self.qcol_sequence = []
        for qual in range(1,self.num_quality_scores+1):
            self.qcol_sequence.append('q'+str(qual))

    def _setup_data(self):
        self.data = { 'lane': [], 'cycle': [], 'tile': [] }
        
        for qual in self.qcol_sequence:
            self.data[qual] = []

    def _setup_read_tiers(self):    
        """
        Concocts a 'read tier' array describing the start-number of cycles per 
        individual Read.
        
        (E.g. read_config of 151/6/151 becomes read_tiers=[151,157,308]
        """ 

        self.read_tiers = []     # list of tuples

        last_tier = 0
        num_reads = len(self.read_config)
    
        for x in range(0,num_reads):
            new_tier = last_tier + self.read_config[x]['cycles'] 
            self.read_tiers.append(new_tier)
            last_tier = new_tier

        dmesg(self.read_tiers, 3)
        
    def get_df_col_sequence(self):
        "returns array of column names in correct order for DataFrame (.df)"
        out = ['cycle', 'lane', 'tile']
        out.extend(self.qcol_sequence)
        return out

    def index_quality(self, target_qscore=30):
        "Convenience method to return index read's % quality for target_qscore (default 30)"
    
        # Assumes only one Index read, which is not a good assumption. 
        # No plans to fix unless specifically requested. (Would probably make more
        # sense just to get rid of this function. -nm) 
        
        for read in self.read_config:
            if read['is_index']:
                return self.get_qscore_percentage(target_qscore, read['read_num']-1)

    def get_qscore_percentage(self, target_qscore=30, read_num=-1):
        """Returns PERCENTAGE of quality scores at or above target_qscore. 
        
        Supplying read_num=-1 returns qscore percentage across all reads. 
        
        :param target_qscore: int designates target quality level (default: 30)
        :param read_num: int specifies read number (default: -1)."
        """
    
        q_upper_cols = [x for x in self.data.keys() if x[0]=='q' and int(x[1:]) > target_qscore-1 ]
        q_upper_df = self.idf[q_upper_cols] 

        if read_num==-1:
            # return %>=qn for entire data set.
            q_upper_sum = q_upper_df.values.sum()
            q_total_sum = self.idf.values.sum()

        else:
            # segment Qscores by read_num. Let IndexError be raised for too-high read_num.
            # read_tiers example: [151,157,308] 
        
            cycle_start = 0 if read_num==0 else self.read_tiers[read_num-1] + 1 
            cycle_end = self.read_tiers[read_num]
        
            tiles = self.flowcell_layout['tilecount']
            lanes = self.flowcell_layout['lanecount']
            surfaces = self.flowcell_layout['surfacecount']

            i_start = 0 if read_num==0 else cycle_start * tiles * lanes * surfaces
            i_end = cycle_end * tiles * lanes * surfaces
        
            q_upper_sum = q_upper_df[i_start:i_end].values.sum()
            q_total_sum = self.idf[i_start:i_end].values.sum()  
        
        # Return a percentage (like in Illumina SAV)
        if q_total_sum:
            return 100 * float(q_upper_sum) / float(q_total_sum)
        else:
            return 0

    def parse_binary(self):
        """ Do the work.  Important: set read_config appropriately, which is
            needed to construct read_tiers to separate Q scores by Read."""

        bs = self.bs   
    
        # QualityMetrics format according to ILMN specs:
        #
        #   byte 0: file version number (4)
        #   byte 1: length of each record
        #   bytes (N * 206 + 2) - (N * 206 + 207): record:
        #       2 bytes: lane number (uint16)
        #       2 bytes: tile number (uint16)
        #       2 bytes: cycle number (uint16)
        #       4 x 50 bytes: number of clusters assigned score (uint32) Q1 through Q50

        self.apparent_file_version = bs.read('uintle:8')
        self.check_version(self.apparent_file_version)

        recordlen = bs.read('uintle:8')  # length of each record

        #read records bytewise per specs in technote_rta_theory_operations.pdf from ILMN
        for i in range(0,((bs.len) / (recordlen * 8))):  # 206 * 8 = 1648 record length in bits
            lane = bs.read('uintle:16')
            tile = bs.read('uintle:16')
            cycle = bs.read('uintle:16')
        
            self.data['lane'].append(lane)
            self.data['tile'].append(tile)
            self.data['cycle'].append(cycle)

            for qual in range(1, self.num_quality_scores + 1):  #(50 entries of 4 bytes each)
                self.data['q'+str(qual)].append(bs.read('uintle:32'))

        self.df = set_column_sequence(pandas.DataFrame(self.data), self.qcol_sequence)
    
        self.idf = self.make_coordinate_plane(self.df, flatten=True)
        
        for read_num in range(self.num_reads):
            q30 = self.get_qscore_percentage(30, read_num)
            q20 = self.get_qscore_percentage(20, read_num)
            self.read_qscore_results['readnum'].append(read_num+1)
            self.read_qscore_results['q30'].append(q30)
            self.read_qscore_results['q20'].append(q20)

    def __str__(self):
        out = ""
        for read in self.read_config:
            out += "Read %i: %f %s\n" % (read['read_num'], self.get_qscore_percentage(30, read['read_num']-1),
                                    "(Index)" if read['is_index'] else "")
        return out

    def to_dict(self, target_qscore=30):
        out = {}
        for read in self.read_config:
            out[read['read_num']] = self.get_qscore_percentage(target_qscore, read['read_num']-1)
        return out


if __name__=='__main__':
    
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print 'supply path to QualityMetrics.bin (or QMetricsOut.bin)'
        sys.exit()
    
    QM = InteropQualityMetrics(filename)
    
    print 'Length of data: %i' % len(QM.data['cycle'])

    print "Quality Score Histogram (all reads)"
    print QM.idf.sum()
    


# -*- coding: utf-8 -*-

import pandas
from bitstring import ReadError

from base_parser_class import InteropBinParser

class InteropIndexMetrics(InteropBinParser):
    "ILMN Quality metrics parser (child class of InteropBinParser)."
    
    __version = 0.1             # version of this parser class.
    supported_versions = [1]    # version(s) of file that this parser supports
    
    # this class doesn't require read_config and flowcell_layout, but 
    # some of the sanity checks it can run on the data do use them.

    def sanity_check( self ):
        "Checks parser output against expectations from parameters of the run."
        #
        # This is the stub of an idea for pan-parser implementation later.
        
        # Does the number of individual records in IndexMetrics jibe w/ flowcell_layout?
        # 
        # Expect: num_records == surfaces * tiles * lanes * indexes  (e.g. MiSeq: 2 x 14 x 1 x 2 = 56)
        assert len(self.data['index_str']) == \
                  self.flowcell_layout['tilecount'] * self.flowcell_layout['surfacecount'] * \
                  len(self.results.keys()) * self.flowcell_layout['lanecount']
                
    def _init_variables( self ):
        self.data = {
            'lane': [], 
            'tile': [], 
            'read': [], 
            'index_str': [], 
            'clusters': [], 
            'name_str': [], 
            'project_str': [] 
            }
            
        self.total_ix_reads_pf = 0  # sum of all index reads passing filter 
        self.results = {}        # after parsing, keyed by unique indexes. 
                                 # value = sum of PF clusters found per unique index.

    def parse_binary(self):
    
        bs = self.bs
    
        # Index Metrics (IndexMetrics.bin and IndexMetricOut.bin) 
        #   Reports the indexes count. Format:
        #   Byte 0: file version (1)
        #   Bytes( variable length): record:
        #   2 bytes: lane number(unint16)
        #   2 bytes: tile number(unint16)
        #   2 bytes: read number(unint16)
        #   2 bytes: number of bytes Y for index name(unint16)
        #   Y bytes: index name string (string in UTF8Encoding)
        #   4 bytes: num of clusters identified as index (uint32)
        #   2 bytes: number of bytes V for sample name(unint16)
        #   V bytes: sample name string (string in UTF8Encoding) 
        #       2 bytes: number of bytes W for sample project(unint16) 
        #   W bytes: sample project string (string in UTF8Encoding)

        self.apparent_file_version = bs.read('uintle:8')  # version number == "1"
        
        self.check_version( self.apparent_file_version )
        
        # Each record is of variable length. Fun!

        try:
            while True:
                self.data['lane'].append(bs.read('uintle:16'))  # lane number
                self.data['tile'].append(bs.read('uintle:16'))  # tile number
                self.data['read'].append(bs.read('uintle:16'))  # read number

                # next 2 bytes: expected index name length in bytes.  
                nextbytes = bs.read('uintle:16')   
                self.data['index_str'].append(bs.read('bytes:%i' % nextbytes) ) #index string
    
                # next 4 bytes: number of clusters identified as index (uint32)
                self.data['clusters'].append(bs.read('uintle:32'))

                # next 2 bytes: expected sample name length in bytes.
                nextbytes = bs.read('uintle:16')
                self.data['name_str'].append(bs.read('bytes:%i' % nextbytes) )      #sample name

                # next 2 bytes: expected sample project string length in bytes.
                nextbytes = bs.read('uintle:16')
                self.data['project_str'].append(bs.read('bytes:%i' % nextbytes))
        
        except ReadError:
            #that's all, folks
            pass

        self.df = pandas.DataFrame( self.data )

        cluster_pivot = self.df.pivot_table('clusters', rows='index_str', aggfunc='sum')
 
        self.total_ix_reads_pf = cluster_pivot.sum()
    
        for ix in self.df.index_str.unique():
            self.results[ix] = cluster_pivot[ix]
        
        self.pivot = self.df.pivot_table('clusters', rows=['index_str', 'project_str', 'name_str'], aggfunc='sum')
        
        # pivot now looks something like this, with any luck:
        """index_str  project_str       name_str                             
            AAACAT     CLIA - WF1265 #1  XL1510-XE2346-LS1430-SQ1000-RE1051-B1    8290786
            CTTGTA     CLIA - WF1265 #1  XL1510-XE2343-LS1429-SQ36-RE1051-A1      8582411
        """    

    def to_dict(self):
        return self.results

    def __str__(self):    
        out = '\n%s\n' % self.pivot
        return out 

if __name__=='__main__':
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print 'supply path to IndexMetrics.bin (or IndexMetricsOut.bin)'
        sys.exit()
    
    IM = InteropIndexMetrics(filename)

    print 'Length of data: %i' % len(IM.data['clusters'])
    print IM




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# InteropDataset
# Library encapsulating the XML and bin files from MiSeq and HiSeq output.
#
# InteropMetadata
# Parser for XML files from MiSeq / HiSeq run data.
#
# See README for intro and basic examples.
#
# March 2013
# by nthmost (naomi.most@invitae.com)
# with lots of help from ECO (eric.olivares@invitae.com)

import time, os
from xml.etree import ElementTree as ET

import pandas

from index_metrics import InteropIndexMetrics
from tile_metrics import InteropTileMetrics
from quality_metrics import InteropQualityMetrics
from error_metrics import InteropErrorMetrics
from corint_metrics import InteropCorrectedIntensityMetrics 
from control_metrics import InteropControlMetrics
from extraction_metrics import InteropExtractionMetrics

from utils import dmesg, select_file_from_aliases

#### CONFIGURABLE THINGS
#
# THIS STUFF WILL BE MOVED INTO A MORE PYTHONIC CONFIG INI TYPE THING.
#
# Configuration of filenames, paths, and types/amounts of output logging.
# (Most people will not need to edit this stuff.)

# Name of directory containing Interop binaries. Usually "InterOp".
BINFILE_DIR_NAME = "InterOp"

# BINARY and XML FILEMAPs
#
# "codenames" (rather than filenames) are used internally in InteropDataset to refer to files.
# (The files themselves contain no explicit indication of what's inside them.)
#
# The FILEMAP variables contain mappings of codename to filename, using aliases.
# to select file, since there are several different filenames out there.
# 
# The first filename to be retrievable will be the one that gets parsed, so view the list as
# a set of decreasing priorities.

BIN_FILEMAP = { 'extraction': ["ExtractionMetricsOut.bin", "ExtractionMetrics.bin"],
                'quality': ["QMetricsOut.bin", "QualityMetricsOut.bin", "QualityMetrics.bin"],
                'error': ["ErrorMetricsOut.bin", "ErrorMetrics.bin"],
                'tile': ["TileMetricsOut.bin", "TileMetrics.bin"],
                'corint': ["CorrectedIntMetricsOut.bin", "CorrectedIntensityMetricsOut.bin", "CorrectedIntMetrics.bin"],
                'control': ["ControlMetricsOut.bin", "ControlMetrics.bin"],
                'image': ["ImageMetricsOut.bin", "ImageMetrics.bin"], 
                'index': ["IndexMetricsOut.bin", "IndexMetrics.bin"] }

XML_FILEMAP = { 'runinfo': ["RunInfo.xml"],
               'runparams': ["runParameters.xml"],
               'reseqstats': ["ResequencingRunStatistics.xml"],
               'completed': ["CompletedJobInfo.xml"] }

#### END OF CONFIGURABLE THINGS ####

class InteropMetadata(object):
    "Parser for sequencer's XML files describing a single run. Supply with directory to instantiate."
    
    __version = 0.1     # version of this parser.

    def __init__(self, xmldir):
        """Takes the absolute path of a sequencing run data directory as sole required variable.
           Attempts to parse CompletedJobInfo.xml (or viable alias). If not available, uses 
           runParameters.xml and/or runInfo.xml, which have some overlapping info (but not all).
           
           2nd param (boolean AUTO, default True) controls whether class attempts to parse automatically.
           Individual parsers can be explicitly called via their respective methods.
           
           Be aware that parsing methods are DESTRUCTIVE to existing instance data."""

        self.xmldir = xmldir
        self.experiment_name = ""        # "RU1453:::/locus/data/run_data//1337/1453"
        self.investigator_name = ""      # "Locus:::Uncle_Jesse - 612 - MiSeq"
        self.runID = ""          # cf CompletedJobInfo.xml / RTARunInfo / Run param "Id"
        self.start_datetime = ""
        self.end_datetime = ""
        self.rta_run_info = { }

        # read_config: a list of dictionaries, each of which describe a single read from the sequencer.
        self.read_config = []
        
        # Flow cell layout: necessary to enable parsing of different machine types' binaries.
        self.flowcell_layout = { }
                
        # Read numbers from ResequencingRunStats.xml 
        # Example: { 'clusters_raw': 19494893, 'clusters_PF': 17381252, 'unindexed': 508055, 'unindexed_PF': 16873197, 
        #               'unaligned': 18572490, 'unaligned_PF': 16973197 }   
        self.resequencing_stats = {}
        
        # CompletedJobInfo.xml is the most complete XML file in the set for what we need,
        # but in a pinch we can parse RunInfo.xml for the essentials.
        #
        # (runParameters.xml can be available but not really usable for several reasons.)
        
        # Getting read_config and flowcell_layout are top priority since they inform the binary parsers.
        #
        # TODO: xml_flex (proposed improvement allowing a config file to set which tokens are required / not required.)

        try:  
            self.parse_CompletedJobInfo(self.get_xml_path('completed'))
        except Exception, e:
            dmesg("[InteropMetadata] Exception: %s" % e, 2)
            self.parse_RunInfo(self.get_xml_path('runinfo'))

        try:
            self.parse_ResequencingRunStats(self.get_xml_path('reseqstats'))
        except Exception, e:
            dmesg("[InteropMetadata] Exception: %s" % e, 2)
            dmesg("[InteropMetadata] resequencing_stats will not be available.", 2)

        try:
            self.parse_RunParameters(self.get_xml_path('runparameters'))
        except Exception, e:
            dmesg("[InteropMetadata] Exception: %s" % e, 2)
            dmesg("[InteropMetadata] runParameters.xml couldn't be parsed; some vars may not be available.", 2)
                
    def get_xml_path(self, codename):
        "returns absolute path to XML file represented by data 'codename'"
        result = select_file_from_aliases(codename, XML_FILEMAP, self.xmldir)
        if result == None:
            raise Exception, "File for codename %s not available" % codename
        else:
            return result
    
    def parse_Run_ET(self, run_ET):
        "parses chunk of XML associated with the RTA Run Info blocks in (at least) 2 xml files."

        self.rta_run_info = { 'instrument': run_ET.find("Instrument").text,     # M00612
                              'flowcell': run_ET.find("Flowcell").text,         # 000000000-A316T
                              'date': run_ET.find("Date").text }                # 130208
    
        flowcell_ET = run_ET.find("FlowcellLayout")
        self.flowcell_layout = { 'lanecount': int(flowcell_ET.attrib['LaneCount']),
                        'surfacecount': int(flowcell_ET.attrib['SurfaceCount']),
                        'swathcount': int(flowcell_ET.attrib['SwathCount']),
                        'tilecount': int(flowcell_ET.attrib['TileCount']) }     
        
        # Run / Reads - describes number of cycles per read (and if read is an Index) in sequencing run.
        
        # Because parsing is understood to be destructive, and Reads can be found in multiple files,
        # we start by emptying out whatever's currently in the read_config array for this instance.
        
        self.read_config = []
        read_num = 0
        for item in run_ET.find("Reads"):
            read_num += 1       # redundant safety assignment to read_num
            self.read_config.append( {'read_num': read_num,
                                      'cycles': int(item.attrib['NumCycles']), 
                                      'is_index': True if item.attrib["IsIndexedRead"]=="Y" else False } )      
    
    def parse_ResequencingRunStats(self, filepath):
        "parses ResequencingRunStatistics.xml (or viable alias) to fill instance variables."
        
        tree = ET.parse(filepath)
        root = tree.getroot()   # should be "StatisticsResequencing"
        runstats_ET = root.find("RunStats")
        
        self.resequencing_stats = { 'clusters_raw': int(runstats_ET.find("NumberOfClustersRaw").text),
                                    'clusters_pf': int(runstats_ET.find("NumberOfClustersPF").text),
                                    'unindexed': int(runstats_ET.find("NumberOfUnindexedClusters").text),
                                    'unindexed_pf': int(runstats_ET.find("NumberOfUnindexedClustersPF").text),
                                    'unaligned': int(runstats_ET.find("NumberOfUnalignedClusters").text),
                                    'unaligned_pf': int(runstats_ET.find("NumberOfUnalignedClustersPF").text),
                                    'duplicate': int(runstats_ET.find("NumberOfDuplicateClusters").text) }

    def parse_CompletedJobInfo(self, filepath):
        "parses CompletedJobInfo.xml (or viable alias) to fill instance variables."
        # comments show example data from a real MiSeq run (2013/02)
        
        tree = ET.parse(filepath)
        root = tree.getroot()       #should be "AnalysisJobInfo"
    
        # Something to be aware of: RTARunInfo contains a "version" attribute.
        # (This parser knows how to deal with version 2.)
        self.rta_version = root.find("RTARunInfo").attrib['Version']

        # original location of data output from the sequencer.
        self.output_folder = root.find("RTAOutputFolder").text 
        
        # TODO: reformat these dates into something nicer (more useable).
        self.start_datetime = root.find("StartTime").text           # 2013-02-09T15:51:50.0811937-08:00
        self.end_datetime = root.find("CompletionTime").text        # 2013-02-09T16:06:44.0124452-08:00
        
        # dechunk all of the major sections we want to extract data from.
        sheet_ET = root.find("Sheet")
        header_ET = sheet_ET.find("Header")
        run_ET = root.find("RTARunInfo").find("Run")
        
        # Sheet / *
        try:
            self.runtype = sheet_ET.find("Type").text       # MiSeq, HiSeq, etc.
        except AttributeError:
            #older (early 2012) XML files have no "Type" token.
            self.runtype = ""
                
        # Sheet / Header / *
        # TODO: xml_flex
        try:
            self.investigator_name = header_ET.find("InvestigatorName").text
            self.project_name = header_ET.find("ProjectName").text
            self.experiment_name = header_ET.find("ExperimentName").text
        except AttributeError:
            pass

        # RTARunInfo / Run / *          
        self.runID = run_ET.attrib["Id"] 

        self.parse_Run_ET(run_ET)

    def parse_RunParameters(self, filepath):
        "partially implemented, not essential. (No FlowcellLayout in this file.)"
        tree = ET.parse(filepath)
        root = tree.getroot()

        # TODO: xml_flex - designate source priority for tokens found in multiple sources. 
        #self.runID = root.find("RunID").text
        
        # TODO: normalize this variable
        #self.start_datetime = root.find('RunStartDate').text    # format: 130208 YYMMDD

        self.experiment_name = root.find('ExperimentName').text
        
        for item in root.find('Reads'):
            #Different format from that in CompletedJobInfo.xml (contains read Number)
            self.read_config.append( {'read_num': int(item.attrib['Number']),
                                      'cycles': int(item.attrib['NumCycles']), 
                                      'is_index': True if item.attrib['IsIndexedRead']=='Y' else False } )


    def parse_RunInfo(self, filepath):
        "parses Reads, Date, Flowcell, Instrument out of runInfo.xml"
        tree = ET.parse(filepath)
        run_ET = tree.getroot().find('Run')     #nothing of use in this file except <Run> subelement. 
        
        self.runID = run_ET.attrib['Id']
        
        #? is this useful information? if so, what for?  (It's not in CompletedJobInfo.xml)
        #self.runNumber = run_ET.attrib['Number']    
        
        self.rta_run_info = self.parse_Run_ET(run_ET)
        
    def prettyprint_read_config(self):
        out = "Read Config:"
        for read in self.read_config:
            out += "    Read %i: %i cycles %s" % (read['read_num'], read['cycles'], 
                                    "(Index)" if read['is_index'] else "")
        return out

    def prettyprint_flowcell_layout(self):
        out = """Flowcell Layout:
        Tiles: %(tilecount)i
        Lanes: %(lanecount)i
        Surfaces: %(surfacecount)i
        Swaths: %(swathcount)i""" % self.flowcell_layout
        return out

    def __str__(self):
        "Print the most important metadata (flowcell layout and read config)"
        out = self.prettyprint_read_config() + "\n" + self.prettyprint_flowcell_layout()
        return out


class InteropDataset(object):
    """Encapsulates the physical files related to this sequencing run. 
       Absolves other classes of having to know about files and directories.
       Combines parser results to emulate Illumina Sequencing Analysis Viewer."""

    directory = ""  # supplied directory name to instantiate class (relative path)
    fullpath = ""   # calculated absolute filesystem path after instantiation
    xmldir = ""     # typically equivalent to fullpath
    bindir = ""     # typically = fullpath/Interop/

    meta = None

    def __init__(self, targetdir):
        "Supply a path (directory) that should contain XML files, with an InterOp directory within it."

        self.directory = targetdir

        self.xmldir = self.directory
        self.bindir = os.path.join(self.directory, BINFILE_DIR_NAME)
        
        # aggregate results of parsing one or more XML files        
        self.meta = self.Metadata()
    
        # holders for the parser objects. Reference via *Metrics() classes.
        self._quality_metrics = None
        self._tile_metrics = None
        self._index_metrics = None
        self._error_metrics = None
        self._corint_metrics = None
        self._extraction_metrics = None
        self._control_metrics = None

    def get_binary_path(self, codename):
        "returns absolute path to binary file represented by data 'codename'"
        path = select_file_from_aliases(codename, BIN_FILEMAP, self.bindir)
        if path is None:
            raise InteropFileNotFoundError("No suitable binary found for {} in directory {}".format(codename, self.bindir))
        else:
            return path
    
    def Metadata(self, reload=False):
        "returns InteropMetadata class generated from this dataset's XML files"
        if self.meta == None or reload == True:
            self.meta = InteropMetadata(self.xmldir)
        return self.meta
    
    def QualityMetrics(self, reload=False):
        "Returns InteropQualityMetrics object from the 'quality' binary in this dataset."
        if self._quality_metrics == None or reload == True:
            self._quality_metrics = InteropQualityMetrics(self.get_binary_path('quality'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._quality_metrics
        
    def TileMetrics(self, reload=False):
        "Returns InteropTileMetrics object from the 'tile' binary in this dataset."
        if self._tile_metrics == None or reload == True:
            self._tile_metrics = InteropTileMetrics(self.get_binary_path('tile'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._tile_metrics

    def IndexMetrics(self, reload=False):
        "Returns InteropIndexMetrics object from the 'index' binary in this dataset."
        if self._index_metrics == None or reload == True:
            self._index_metrics = InteropIndexMetrics(self.get_binary_path('index'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._index_metrics

    def ControlMetrics(self, reload=False):
        "Returns InteropControlMetrics object from the 'control' binary in this dataset."
        if self._control_metrics == None or reload == True:
            self._control_metrics = InteropControlMetrics(self.get_binary_path('control'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._control_metrics

    def ErrorMetrics(self, reload=False):
        "Returns InteropErrorMetrics object from the 'error' binary in this dataset."
        if self._error_metrics == None or reload == True:
            self._error_metrics = InteropErrorMetrics(self.get_binary_path('error'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._error_metrics

    def ExtractionMetrics(self, reload=False):
        "Returns InteropExtractionMetrics object from the 'extraction' binary in this dataset."
        if self._extraction_metrics == None or reload == True:
            self._extraction_metrics = InteropExtractionMetrics(self.get_binary_path('extraction'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._extraction_metrics

    def CorrectedIntensityMetrics(self, reload=False):
        "Returns InteropCorrectedIntensityMetrics object from the 'extraction' binary in this dataset."
        if self._error_metrics == None or reload == True:
            self._corint_metrics = InteropCorrectedIntensityMetrics(self.get_binary_path('corint'), 
                                    flowcell_layout=self.meta.flowcell_layout,
                                    read_config=self.meta.read_config )
        return self._corint_metrics

#TODO: ImageMetrics
#    def ImageMetrics(self, reload=False):
#        "returns InteropImageMetrics object from the 'image' binary in this dataset"
#        if self._image_metrics == None or reload == True:
#            self._image_metrics = InteropImageMetrics(self.get_binary_path('image'), 
#                                    flowcell_layout=self.meta.flowcell_layout,
#                                    read_config=self.meta.read_config )
#        return self._image_metrics
    
  
## Custom Exceptions for InteropDataset

class InteropFileNotFoundError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)

        
## Command Line helper functions below
def print_sample_dataset(ID):
    meta = ID.Metadata()

    print ""
    print "METADATA"
    print "--------"
    print meta
    print ""
    
    print "Resequencing Statistics:" 
    print meta.resequencing_stats
    print ""
    
    tm = ID.TileMetrics()
    print "SUMMARY"
    print "-------"
    print "(tile metrics)\n"
    print tm
    print ""
    
    # QualityMetrics usually take a while...
    print "QUALITY"
    print "-------"
    qm = ID.QualityMetrics()
    print "(% >= Q30 per read)\n"
    print qm
    print ""

    im = ID.IndexMetrics()
    print "INDEXING"
    print "--------\n"
    print "Total Reads: %i" % tm.num_clusters
    print "Reads PF: %i" % tm.num_clusters_pf
    print "Percentage Reads Identified (PF): %f" % (float(im.total_ix_reads_pf / tm.num_clusters_pf)*100)
    print ""
    print im
    print ""

    print "ERRORS"
    print "------"
    try:
        em = ID.ErrorMetrics()
        print "(sum of all types of errors across all reads)"
        idf = em.make_coordinate_plane(em.df)
        print idf.sum()
    except TypeError:
        print "None. (no error metrics binary in this dataset.)"
    finally:
        print ""

    """
    print "INTENSITY"
    print "---------"
    try:
        cm = ID.CorrectedIntensityMetrics()
        print "(sample of raw data)"
        if cm is not None:
            print cm.idf.head()
        else:
            print "CorrectedIntensityMetrics parsing failed."
    except TypeError:
        print "No CorrectedIntensityMetrics binary in this dataset."
    finally:
        print ""
    """

if __name__=='__main__':
    import sys

    try:
        dirname = sys.argv[1] 
    except IndexError:
        print "Supply the absolute or relative path to a directory of sequencing rundata."
        print "Example:  python interop.py /home/username/seqruns/2013-02/0"
        sys.exit()
        
    myDataset = InteropDataset(dirname)

    #print myDataset.meta.read_config

    print_sample_dataset(myDataset)
    



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta
import pandas

from base_parser_class import InteropBinParser

anno_domini = datetime(1, 1, 1)

class InteropExtractionMetrics(InteropBinParser):

    __version = 0.1
    supported_versions = [2]

    def _init_variables(self):
        self.data = { 
                    'lane': [], 
                    'tile': [], 
                    'cycle': [], 
                    'fwhm_A': [], 
                    'fwhm_C': [], 
                    'fwhm_G': [], 
                    'fwhm_T': [], 
                    'intensity_A': [], 
                    'intensity_C': [], 
                    'intensity_G': [], 
                    'intensity_T': [],
                    'datetime': [],
                    }
        
    def parse_binary(self):
        bs = self.bs
        
        # Extraction Metrics (ExtractionMetricsOut.bin)
        # Contains extraction metrics such as fwhm (full width at half maximum) scores and raw intensities
        # Format:
        #   byte 0: file version number (2)
        #   byte 1: length of each record
        #   bytes (N * 38 + 2) - (N *38 + 39): record:
        #     2 bytes: lane number (uint16)
        #     2 bytes: tile number (uint16)
        #     2 bytes: cycle number (uint16)
        #     4 x 4 bytes: fwhm scores (float) for channel [A, C, G, T] respectively 
        #     2 x 4 bytes: intensities (uint16) for channel [A, C, G, T] respectively 
        #     8 bytes: date/time of CIF creation --> serialized C# datetime object 
        #   ...Where N is the record index

        self.apparent_file_version = bs.read('uintle:8')
        self.check_version(self.apparent_file_version)
        
        recordlen = bs.read('uintle:8')  # length of each record

        for i in range(0,((bs.len) / (recordlen * 8))):  # record length in bits
            self.data['lane'].append(bs.read('uintle:16'))
            self.data['tile'].append(bs.read('uintle:16'))
            self.data['cycle'].append(bs.read('uintle:16'))
    
            # 4 x 4 bytes: fwhm scores (float) for channel [A, C, G, T] respectively 
            self.data['fwhm_A'].append(bs.read('floatle:32'))    
            self.data['fwhm_C'].append(bs.read('floatle:32'))
            self.data['fwhm_G'].append(bs.read('floatle:32'))
            self.data['fwhm_T'].append(bs.read('floatle:32'))
    
            # 2 x 4 bytes: intensities (uint16) for channel [A, C, G, T] respectively 
            self.data['intensity_A'].append(bs.read('uintle:16'))
            self.data['intensity_C'].append(bs.read('uintle:16'))
            self.data['intensity_G'].append(bs.read('uintle:16'))
            self.data['intensity_T'].append(bs.read('uintle:16'))
    
            # 8 bytes: date/time of CIF creation
            datetime_bits = bs.read(64)
            # first 2 bits of last byte represent "kind" of date
            # we don't care about "kind", so let's zero those bits
            datetime_bits.set(0, (56, 57))
            # the rest is a 62bit integer giving 100ns since midnight Jan 1, 0001
            microseconds = timedelta(microseconds=datetime_bits.uintle / 10)
            self.data['datetime'].append(anno_domini + microseconds)

        self.df = pandas.DataFrame(self.data)

        #self.idf = self.make_coordinate_plane(self.df)

    def __str__(self): 
        #TODO: better printout
        out = '%s' % self.df.head()
        return out

if __name__=='__main__':
    import sys
    
    try:
        filename = sys.argv[1]
    except:
        print 'supply path to ExtractionMetrics.bin'
        sys.exit()
    
    EM = InteropExtractionMetrics(filename)
    print EM



from setuptools import setup, find_packages

setup(
    name = "illuminate",
    version = "0.5.4",
    description = "Analytics toolkit for Illumina sequencer metrics.",
    url="https://bitbucket.org/invitae/illuminate",
    author = "InVitae Inc.",
    author_email = "developers@invitae.com",
    maintainer = "Naomi Most",
    maintainer_email = "naomi.most@invitae.com",
    license = "MIT",
    zip_safe = False,
    packages = find_packages(),
    install_requires = ["bitstring>=3.1.0",
                        "docopt",
                        "numpy>=1.6.2",
                        "pandas>=0.11",
                        ],
    )



from docopt import docopt
from interop import InteropDataset

__doc__="""ILLUMINATE

Usage: illuminate [options] <datapath>...

  -h --help             Show this screen.
  --version             Show version.
  -v, --verbose         Increase verbosity           
  -q, --quiet           Suppress all console output   
  -d, --dump=<outfile>  Output parser results to file. 
  -i, --interactive     Load dataset into iPython for interactive fun.

  --all             Parse and print everything

  --meta            Print metadata
  --quality         Parse quality metrics
  --tile            Parse tile metrics
  --index           Parse index metrics
  --error           Parse error metrics
  --corint          Parse corrected intensity metrics
  --extraction      Parse extraction metrics
  --control         Parse control metrics
  --image           Parse image metrics (not yet supported)

  --csv=<csv_filename> Output results as CSV (not yet supported)
"""

#TODO: SAV_emu
"""
  --analysis        Print an emulation of the Analysis screen
  --summary         Print an emulation of the Summary screen
  --indexing        Print an emulation of the Indexing screen
"""

VERBOSITY = 1
DEBUG = False
OUTFILE = None

def dmesg(msg, lvl=1):
    #msg = '[%f] ' % time.time() + msg
    if DEBUG: msg = '[DEBUG] ' + msg
    if VERBOSITY >= lvl:
        print msg
    if OUTFILE:
        OUTFILE.write(msg+'\n')
        OUTFILE.flush()

def calculate_verbosity(verbose, quiet):
    global VERBOSITY
    if quiet:
        VERBOSITY=0
    elif verbose and not quiet:
        VERBOSITY=2
    else:
        VERBOSITY=1

def arrange_writing_to_file(filename):
    global OUTFILE
    if filename:
        try:
            OUTFILE = open(filename, 'w')
        except Exception, e:
            dmesg('%s' % e, 1)

def run_metrics_object(InteropObject, title):
    dmesg(title, 1)
    dmesg('-' * len(title), 1)
    dmesg('%s' % InteropObject, 1)

if __name__=='__main__':

    args = docopt(__doc__,version='0.4')

    if args['--interactive']:
        from IPython import embed
        myDataset = InteropDataset(args['<datapath>'][0]) 
        embed()
    else:
        calculate_verbosity(args['--verbose'], args['--quiet'])
        arrange_writing_to_file(args['--dump'])

        #TODO: print nicer error messages when binary file is missing

        for datapath in args['<datapath>']:
            ID = InteropDataset(datapath)
            if args['--all'] or args['--meta']:
                run_metrics_object(ID.meta, "METADATA")
            if args['--all'] or args['--tile']:
                run_metrics_object(ID.TileMetrics(), "SUMMARY")
            if args['--all'] or args['--quality']:
                run_metrics_object(ID.QualityMetrics(), "QUALITY")
            if args['--all'] or args['--index']:
                run_metrics_object(ID.IndexMetrics(), "INDEXING")
            if args['--all'] or args['--error']:
                run_metrics_object(ID.ErrorMetrics(), "ERRORS")
            if args['--all'] or args['--corint']:
                run_metrics_object(ID.CorrectedIntensityMetrics(), "INTENSITY")
            if args['--all'] or args['--extraction']:
                run_metrics_object(ID.ExtractionMetrics(), "EXTRACTION")
            if args['--all'] or args['--control']:
                run_metrics_object(ID.ControlMetrics(), "CONTROL")

        #TODO: SAV_emu
    if OUTFILE:
        OUTFILE.close()

